@@ -6,6 +6,7 @@ import hashlib
 import itertools
 import os
 import pathlib
+import platform
 import pprint
 import shutil
 import tempfile
@@ -17,33 +18,35 @@ from . import version
 
 
 # Template for patching __init__.py so that the vendored DLLs are loaded at
-# runtime. An empty triple-quoted string is placed at the beginning so that the
-# comment "start delvewheel patch" does not show up when the built-in help
-# system help() is invoked on the package. For Python >=3.8, we use the
-# os.add_dll_directory() function so that the folder containing the vendored
-# DLLs is added to the DLL search path. For Python 3.7 or lower, this function
-# is unavailable, so we preload the DLLs. Whenever Python needs a vendored DLL,
-# it will use the already-loaded DLL instead of searching for it.
+# runtime. If the patch would be placed at the beginning of the file, an empty
+# triple-quoted string is placed at the beginning so that the comment
+# "start delvewheel patch" does not show up when the built-in help system help()
+# is invoked on the package. For Python >=3.8, we use the os.add_dll_directory()
+# function so that the folder containing the vendored DLLs is added to the DLL
+# search path. For Python 3.7 or lower, this function is unavailable, so we
+# preload the DLLs. Whenever Python needs a vendored DLL, it will use the
+# already-loaded DLL instead of searching for it.
 #
-# To use the template, call str.format(), passing in an identifying string, the
-# name of the directory containing the vendored DLLs, and the name of the file
-# containing the DLL load order.
+# To use the template, call str.format(), passing in
+# 0. '""""""' if the patch would be at the start of the file else ''
+# 1. an identifying string such as the delvewheel version
+# 2. the name of the directory containing the vendored DLLs
+# 3. the name of the file containing the DLL load order.
 _patch_init_template = """
 
-""\"""\"  # start delvewheel patch
-def _delvewheel_init_patch_{0}():
+{0}# start delvewheel patch
+def _delvewheel_init_patch_{1}():
     import os
     import sys
-    libs_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, {1!r}))
+    libs_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, {2!r}))
     if sys.version_info[:2] >= (3, 8):
-        if os.path.exists(os.path.join(sys.base_prefix, 'conda-meta')):
+        conda_workaround = sys.version_info[:3] < (3, 9, 9) and os.path.exists(os.path.join(sys.base_prefix, 'conda-meta'))
+        if conda_workaround:
             # backup the state of the environment variable CONDA_DLL_SEARCH_MODIFICATION_ENABLE
             conda_dll_search_modification_enable = os.environ.get('CONDA_DLL_SEARCH_MODIFICATION_ENABLE')
             os.environ['CONDA_DLL_SEARCH_MODIFICATION_ENABLE'] = '1'
-
         os.add_dll_directory(libs_dir)
-
-        if os.path.exists(os.path.join(sys.base_prefix, 'conda-meta')):
+        if conda_workaround:
             # restore the state of the environment variable CONDA_DLL_SEARCH_MODIFICATION_ENABLE
             if conda_dll_search_modification_enable is None:
                 os.environ.pop('CONDA_DLL_SEARCH_MODIFICATION_ENABLE', None)
@@ -51,14 +54,14 @@ def _delvewheel_init_patch_{0}():
                 os.environ['CONDA_DLL_SEARCH_MODIFICATION_ENABLE'] = conda_dll_search_modification_enable
     else:
         from ctypes import WinDLL
-        with open(os.path.join(libs_dir, {2!r})) as file:
+        with open(os.path.join(libs_dir, {3!r})) as file:
             load_order = file.read().split()
         for lib in load_order:
             WinDLL(os.path.join(libs_dir, lib))
 
 
-_delvewheel_init_patch_{0}()
-del _delvewheel_init_patch_{0}
+_delvewheel_init_patch_{1}()
+del _delvewheel_init_patch_{1}
 # end delvewheel patch
 
 """
@@ -178,8 +181,6 @@ class WheelRepair:
         libs_dir is the name of the directory where DLLs are stored."""
         print(f'patching {os.path.relpath(init_path, self._extract_dir)}')
 
-        patch_init_contents = _patch_init_template.format(version.__version__.replace('.', '_'), libs_dir, load_order_filename)
-
         open(init_path, 'a+').close()
         with open(init_path) as file:
             init_contents = file.read()
@@ -195,6 +196,7 @@ class WheelRepair:
 
         if future_import_lineno > 0:
             # insert patch after the last __future__ import
+            patch_init_contents = _patch_init_template.format('', version.__version__.replace('.', '_'), libs_dir, load_order_filename)
             init_contents_split = init_contents.splitlines(True)
             with open(init_path, 'w') as file:
                 file.write(''.join(init_contents_split[:future_import_lineno]))
@@ -203,12 +205,15 @@ class WheelRepair:
                 file.write(''.join(init_contents_split[future_import_lineno:]))
         elif docstring is None:
             # prepend patch
+            patch_init_contents = _patch_init_template.format('""""""', version.__version__.replace('.', '_'), libs_dir, load_order_filename)
             with open(init_path, 'w') as file:
                 file.write(patch_init_contents)
                 file.write(init_contents)
         else:
-            # verify that the first child node is the docstring
+            # place patch just after docstring
+            patch_init_contents = _patch_init_template.format('', version.__version__.replace('.', '_'), libs_dir, load_order_filename)
             if len(children) == 0 or not isinstance(children[0], ast.Expr) or ast.literal_eval(children[0].value) != docstring:
+                # verify that the first child node is the docstring
                 raise ValueError('Error parsing __init__.py: docstring exists but is not the first element of the parse tree')
             if len(children) == 1:
                 # append patch
@@ -281,22 +286,20 @@ class WheelRepair:
                 return file.read().strip()
         return ''
 
-    def _split_dependency_paths(self, dependency_paths: typing.Iterable):
+    def _split_dependency_paths(self, dependency_paths: typing.Iterable) -> typing.Tuple[typing.Set, typing.Set]:
         """Given an iterable of DLL paths, partition the contents into a tuple
-        of sorted lists
+        of sets
         (dependency_paths_in_wheel, dependency_paths_outside_wheel).
         dependency_paths_in_wheel contains the paths to DLLs that are already in
         the wheel, and dependency_paths_outside_wheel contains the paths to DLLs
         that are not in the wheel."""
-        dependency_paths_in_wheel = []
-        dependency_paths_outside_wheel = []
+        dependency_paths_in_wheel = set()
+        dependency_paths_outside_wheel = set()
         for dependency_path in dependency_paths:
             if pathlib.Path(self._extract_dir) in pathlib.Path(dependency_path).parents:
-                dependency_paths_in_wheel.append(dependency_path)
+                dependency_paths_in_wheel.add(dependency_path)
             else:
-                dependency_paths_outside_wheel.append(dependency_path)
-        dependency_paths_in_wheel.sort()
-        dependency_paths_outside_wheel.sort()
+                dependency_paths_outside_wheel.add(dependency_path)
         return dependency_paths_in_wheel, dependency_paths_outside_wheel
 
     def show(self) -> None:
@@ -319,16 +322,34 @@ class WheelRepair:
                 if filename.endswith('.pyd'):
                     extension_module_path = os.path.join(root, filename)
                     extension_module_paths.append(extension_module_path)
-                    discovered, ignored, not_found = patch_dll.get_all_needed(extension_module_path, self._add_dlls, self._no_dlls, self._wheel_dirs, 'ignore')
+                    discovered, ignored, not_found = patch_dll.get_all_needed(extension_module_path, self._no_dlls, self._wheel_dirs, 'ignore')
                     dependency_paths |= discovered
                     ignored_dll_names |= ignored
                     not_found_dll_names |= not_found
 
+        # find extra dependencies specified with --add-dll
+        interpreter_bitness = 64 if platform.architecture()[0] == '64bit' else 32
+        extra_dependency_paths = set()
+        for dll_name in self._add_dlls:
+            path = patch_dll.find_library(dll_name, None, interpreter_bitness)
+            if path:
+                extra_dependency_paths.add(path)
+            else:
+                not_found_dll_names.add(dll_name)
+
         if self._ignore_in_wheel:
             dependency_paths_in_wheel, dependency_paths_outside_wheel = self._split_dependency_paths(dependency_paths)
+            for path in dependency_paths_in_wheel.copy():
+                if os.path.basename(path).lower() in self._add_dlls:
+                    dependency_paths_in_wheel.remove(path)
+            dependency_paths_in_wheel = list(dependency_paths_in_wheel)
+            dependency_paths_in_wheel.sort()
+            dependency_paths_outside_wheel |= extra_dependency_paths
+            dependency_paths_outside_wheel = list(dependency_paths_outside_wheel)
+            dependency_paths_outside_wheel.sort()
         else:
             dependency_paths_in_wheel = None
-            dependency_paths_outside_wheel = list(dependency_paths)
+            dependency_paths_outside_wheel = list(dependency_paths | extra_dependency_paths)
             dependency_paths_outside_wheel.sort()
         print('The following DLLs will be copied into the wheel.')
         if dependency_paths_outside_wheel:
@@ -347,6 +368,7 @@ class WheelRepair:
             else:
                 print('    None')
 
+        ignored_dll_names -= {os.path.basename(p).lower() for p in dependency_paths_outside_wheel}
         ignored_dll_names = list(ignored_dll_names)
         ignored_dll_names.sort()
         print("\nThe following DLLs are assumed to be present in the end user's environment and will not be copied into the wheel.")
@@ -372,7 +394,7 @@ class WheelRepair:
             print(f'Delvewheel {repair_version} has already repaired this wheel.')
             return
 
-        # find dependencies and copy them into wheel
+        # find dependencies
         print('finding DLL dependencies')
         dependency_paths = set()
         ignored_dll_names = set()
@@ -389,14 +411,44 @@ class WheelRepair:
                     elif self._verbose >= 1:
                         print(f'analyzing package-level extension module {os.path.relpath(extension_module_path, self._extract_dir)}')
                     extension_module_paths.append(extension_module_path)
-                    discovered, ignored = patch_dll.get_all_needed(extension_module_path, self._add_dlls, self._no_dlls, self._wheel_dirs)[:2]
+                    discovered, ignored = patch_dll.get_all_needed(extension_module_path, self._no_dlls, self._wheel_dirs)[:2]
                     dependency_paths |= discovered
                     ignored_dll_names |= ignored
-        if not dependency_paths:
+
+        # if --ignore-in-wheel is specified, ignore DLLs that were found inside
+        # the wheel unless they are specified with --add-dll
+        if self._ignore_in_wheel:
+            dependency_paths_in_wheel, dependency_paths_outside_wheel = self._split_dependency_paths(dependency_paths)
+            for p in dependency_paths_in_wheel:
+                name_lower = os.path.basename(p).lower()
+                no_mangles.add(name_lower)
+                no_mangles.update(patch_dll.get_direct_mangleable_needed(p, self._no_dlls, no_mangles))
+                if name_lower not in self._add_dlls:
+                    ignored_dll_names.add(name_lower)
+            dependency_paths = dependency_paths_outside_wheel
+
+        # find extra dependencies specified with --add-dll that have not yet
+        # been found
+        dependency_names = {os.path.basename(p) for p in dependency_paths}  # this is NOT lowercased
+        dependency_names_lower = {name.lower() for name in dependency_names}
+        interpreter_bitness = 64 if platform.architecture()[0] == '64bit' else 32
+        extra_dependency_paths = set()
+        for dll_name in self._add_dlls:
+            if dll_name in dependency_names_lower:
+                continue
+            path = patch_dll.find_library(dll_name, None, interpreter_bitness)
+            if path:
+                extra_dependency_paths.add(path)
+            else:
+                raise FileNotFoundError(f'{dll_name} not found')
+
+        if not dependency_paths and not extra_dependency_paths:
             print('no external dependencies are needed')
             return
         if self._verbose >= 1:
-            print(f'External dependencies to copy into the wheel are\n{pp.pformat(set(os.path.basename(p) for p in dependency_paths))}')
+            to_copy = set(os.path.basename(p) for p in dependency_paths | extra_dependency_paths)
+            ignored_dll_names -= {name.lower() for name in to_copy}
+            print(f'External dependencies to copy into the wheel are\n{pp.pformat(to_copy)}')
             print(f'External dependencies not to copy into the wheel are\n{pp.pformat(ignored_dll_names)}')
         if top_level_ext_module_names:
             # At least 1 extension module is top-level, so we cannot use
@@ -411,25 +463,25 @@ class WheelRepair:
             libs_dir = os.path.join(self._extract_dir, libs_dir_name)
         os.makedirs(libs_dir, exist_ok=True)
         print(f'copying DLLs into {os.path.relpath(libs_dir, self._extract_dir)}')
-        for dependency_path in dependency_paths:
+        for dependency_path in dependency_paths | extra_dependency_paths:
             if self._verbose >= 1:
                 print(f'copying {dependency_path} -> {os.path.join(libs_dir, os.path.basename(dependency_path))}')
             shutil.copy2(dependency_path, libs_dir)
 
         # mangle library names
-        dependency_names = {os.path.basename(p) for p in dependency_paths}
-        name_mangler = {}  # dict from old name to new name
+        name_mangler = {}  # dict from lowercased old name to new name
         if no_mangle_all:
             print('skip mangling DLL names')
         else:
             print('mangling DLL names')
             for lib_name in dependency_names:
-                if not any(lib_name.startswith(prefix) for prefix in dll_list.no_mangle_prefixes) and \
-                        lib_name not in no_mangles:
+                # lib_name is NOT lowercased
+                if not any(lib_name.lower().startswith(prefix) for prefix in dll_list.no_mangle_prefixes) and \
+                        lib_name.lower() not in no_mangles:
                     root, ext = os.path.splitext(lib_name)
                     with open(os.path.join(libs_dir, lib_name), 'rb') as lib_file:
                         root = f'{root}-{self._hashfile(lib_file)}'
-                    name_mangler[lib_name] = root + ext
+                    name_mangler[lib_name.lower()] = root + ext
             for extension_module_path in extension_module_paths:
                 extension_module_name = os.path.basename(extension_module_path)
                 if self._verbose >= 1:
@@ -437,16 +489,17 @@ class WheelRepair:
                 needed = patch_dll.get_direct_mangleable_needed(extension_module_path, self._no_dlls, no_mangles)
                 patch_dll.replace_needed(extension_module_path, needed, name_mangler)
             for lib_name in dependency_names:
+                # lib_name is NOT lowercased
                 if self._verbose >= 1:
-                    if lib_name in name_mangler:
-                        print(f'repairing {lib_name} -> {name_mangler[lib_name]}')
+                    if lib_name.lower() in name_mangler:
+                        print(f'repairing {lib_name} -> {name_mangler[lib_name.lower()]}')
                     else:
                         print(f'repairing {lib_name} -> {lib_name}')
                 lib_path = os.path.join(libs_dir, lib_name)
                 needed = patch_dll.get_direct_mangleable_needed(lib_path, self._no_dlls, no_mangles)
                 patch_dll.replace_needed(lib_path, needed, name_mangler)
-                if lib_name in name_mangler:
-                    os.rename(lib_path, os.path.join(libs_dir, name_mangler[lib_name]))
+                if lib_name.lower() in name_mangler:
+                    os.rename(lib_path, os.path.join(libs_dir, name_mangler[lib_name.lower()]))
 
         # Perform topological sort to determine the order that DLLs must be
         # loaded at runtime. We first construct a directed graph where the
@@ -457,11 +510,17 @@ class WheelRepair:
         # loaded.
         print('calculating DLL load order')
         for dependency_name in dependency_names.copy():
-            if dependency_name in name_mangler:
+            # dependency_name is NOT lowercased
+            if dependency_name.lower() in name_mangler:
                 dependency_names.remove(dependency_name)
-                dependency_names.add(name_mangler[dependency_name])
-        graph = {}  # map each DLL to a set of its vendored direct dependencies
+                dependency_names.add(name_mangler[dependency_name.lower()])
+
+        # map from lowercased DLL name to its original case
+        dependency_name_casemap = {dependency_name.lower(): dependency_name for dependency_name in dependency_names}
+
+        graph = {}  # map each lowercased DLL name to a lowercased set of its vendored direct dependencies
         for dll_name in dependency_names:
+            # dll_name is NOT lowercased
             dll_path = os.path.join(libs_dir, dll_name)
             # In this context, delay-loaded DLL dependencies are not true
             # dependencies because they are not necessary to get the DLL to load
@@ -469,14 +528,14 @@ class WheelRepair:
             # were were to consider delay-loaded DLLs as true dependencies.
             # For example, concrt140.dll lists msvcp140.dll in its import table,
             # while msvcp140.dll lists concrt140.dll in its delay import table.
-            graph[dll_name] = patch_dll.get_direct_needed(dll_path, False) & dependency_names
+            graph[dll_name.lower()] = patch_dll.get_direct_needed(dll_path, False) & set(dependency_name_casemap.keys())
         rev_dll_load_order = []
-        no_incoming_edge = {dll_name for dll_name in dependency_names if not any(dll_name in value for value in graph.values())}
+        no_incoming_edge = {dll_name_lower for dll_name_lower in dependency_name_casemap.keys() if not any(dll_name_lower in value for value in graph.values())}
         while no_incoming_edge:
-            dll_name = no_incoming_edge.pop()
-            rev_dll_load_order.append(dll_name)
-            while graph[dll_name]:
-                dependent_dll_name = graph[dll_name].pop()
+            dll_name_lower = no_incoming_edge.pop()
+            rev_dll_load_order.append(dependency_name_casemap[dll_name_lower])
+            while graph[dll_name_lower]:
+                dependent_dll_name = graph[dll_name_lower].pop()
                 if not any(dependent_dll_name in value for value in graph.values()):
                     no_incoming_edge.add(dependent_dll_name)
         if any(graph.values()):
